use std::{
    fmt::{Debug, Display},
    sync::Arc,
};

use anyhow::{anyhow, Result};
use bytes::Bytes;
use futures::Stream;
use tokio::sync::mpsc;
use tokio_stream::StreamExt;

const PREFIX: &'static str = "agents.v1";

#[derive(Debug)]
pub enum Subject {
    Health,
    Exec,
    OpenTerminal,
}

impl Display for Subject {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Subject::Health => write!(f, "health"),
            Subject::Exec => write!(f, "exec"),
            Subject::OpenTerminal => write!(f, "open_terminal"),
        }
    }
}

impl TryFrom<&str> for Subject {
    type Error = anyhow::Error;

    fn try_from(value: &str) -> Result<Self> {
        Ok(match value {
            "health" => Subject::Health,
            "exec" => Subject::Exec,
            "open_terminal" => Subject::OpenTerminal,
            _ => return Err(anyhow!("unknown subject '{}'", value)),
        })
    }
}

#[derive(Debug)]
pub struct Message {
    subject: Subject,
    payload: Bytes,
    // reply: Option<async_nats::Subject>,
}

impl Message {
    fn from_transport(subject: &str, payload: Bytes) -> Result<Self> {
        let subject = subject.try_into()?;

        // msg.headers.unwrap().get("x-idempotent-key");

        Ok(Message {
            subject,
            payload,
            // reply: None,
        })
    }

    pub fn subject(&self) -> &Subject {
        &self.subject
    }

    pub fn body(&self) -> Bytes {
        self.payload.clone()
    }

    pub fn health(health: crate::health::Health) -> Result<Message> {
        Ok(Message {
            subject: Subject::Health,
            payload: Bytes::from_iter(serde_json::to_vec(&health)?),
            // reply: None,
        })
    }
}

#[derive(Clone)]
pub struct Client {
    id: Arc<String>,
    nats: async_nats::Client,
    subj_prefix: Arc<String>,
}

impl Client {
    pub async fn connect(id: &str) -> Result<Self> {
        let nats = async_nats::ConnectOptions::new()
            .name(format!("Prymn Agent {id}"))
            .custom_inbox_prefix(format!("_INBOX_{id}"))
            .user_and_password("demo_agent".to_owned(), "demo_agent_password".to_owned())
            .connect("localhost")
            .await?;

        tracing::debug!("connected to NATS");

        Ok(Self {
            nats,
            id: Arc::new(id.to_owned()),
            subj_prefix: Arc::new(format!("{}.{}", PREFIX, id)),
        })
    }

    pub async fn publish(&self, msg: Message) {
        let subject = format!("agents.v1.{}.{}", self.id, msg.subject);
        self.nats.publish(subject, msg.payload).await.unwrap();
    }

    // pub async fn reply(&self, msg: Message, mut stream: impl Stream<Item = Bytes> + Unpin) {
    //     match msg.reply {
    //         Some(reply) => {
    //             while let Some(data) = stream.next().await {
    //                 self.nats.publish(reply.clone(), data.into()).await.unwrap();
    //             }
    //         }
    //         None => tracing::warn!(?msg, "tried to reply to message without a reply subject"),
    //     }
    // }

    pub async fn messages_channel(&self) -> Result<mpsc::Receiver<Message>> {
        let (sender, receiver) = mpsc::channel(100);
        let mut stream = self.subscribe("*").await?;

        let self_clone = self.clone();
        tokio::spawn(async move {
            while let Some(msg) = stream.next().await {
                self_clone
                    .clone()
                    .open_subchannels(&msg, sender.clone())
                    .await;

                sender.send(msg).await.unwrap();
            }
        });

        Ok(receiver)
    }

    async fn subscribe(&self, subject: &str) -> Result<impl Stream<Item = Message>> {
        let prefix = Arc::clone(&self.subj_prefix);

        let stream = self
            .nats
            .subscribe(format!("{}.{}", self.subj_prefix, subject))
            .await?
            .filter_map(move |data| {
                let subject = data.subject[..].trim_start_matches(&*prefix);

                match Message::from_transport(subject, data.payload) {
                    Ok(msg) => Some(msg),
                    Err(err) => {
                        tracing::warn!("{}", err);
                        None
                    }
                }
            });

        tracing::debug!("subscribed on {}", subject);

        Ok(stream)
    }

    async fn open_subchannels(&self, message: &Message, sender: mpsc::Sender<Message>) {
        fn send_messages_from_stream(
            mut stream: impl Stream<Item = Message> + Send + Unpin + 'static,
            sender: mpsc::Sender<Message>,
        ) {
            tokio::spawn(async move {
                while let Some(message) = stream.next().await {
                    sender.send(message).await.unwrap();
                }
            });
        }

        match message.subject {
            Subject::OpenTerminal => {
                let input_stream = self.subscribe("terminal.key.input").await.unwrap();
                let resize_stream = self.subscribe("terminal.key.resize").await.unwrap();

                send_messages_from_stream(input_stream, sender.clone());
                send_messages_from_stream(resize_stream, sender);
            }
            _ => {}
        }
    }
}

impl Debug for Client {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Client {{ id: {} }}", self.id)
    }
}
